from tensordict import TensorDict
from torch import nn

from rl4co.utils.lightning import get_lightning_device


class REINFORCE(nn.Module):
    """Base model for REINFORCE-based models

    Args:
        env: TorchRL Environment
        policy: Policy (set up in model)
        baseline: REINFORCE Baseline (set up in model)
    """

    def __init__(self, env, policy=None, baseline=None):
        super(REINFORCE, self).__init__()
        self.env = env

    def forward(self, td: TensorDict, phase: str = "train", extra=None, **policy_kwargs):
        # Evaluate model, get costs and log probabilities
        out = self.policy(td, phase, **policy_kwargs)

        if phase == "train":
            # REINFORCE loss: we consider the rewards instead of costs to be consistent with the literature
            bl_val, bl_neg_loss = (
                self.baseline.eval(td, out["reward"]) if extra is None else (extra, 0)
            )
            advantage = out["reward"] - bl_val  # advantage = reward - baseline
            reinforce_loss = -(advantage * out["log_likelihood"]).mean()
            loss = reinforce_loss - bl_neg_loss

            out.update(
                {
                    "loss": loss,
                    "reinforce_loss": reinforce_loss,
                    "bl_loss": -bl_neg_loss,
                    "bl_val": bl_val,
                }
            )

        return out

    def setup(self, lit_module):
        # Make baseline taking model itself and train_dataloader from model as input
        self.baseline.setup(
            self.policy,
            self.env,
            batch_size=lit_module.val_batch_size,
            device=get_lightning_device(lit_module),
            dataset_size=lit_module.cfg.data.val_size,
        )

    def on_train_epoch_end(self, lit_module):
        self.baseline.epoch_callback(
            self.policy,
            env=self.env,
            batch_size=lit_module.val_batch_size,
            device=get_lightning_device(lit_module),
            epoch=lit_module.current_epoch,
            dataset_size=lit_module.cfg.data.val_size,
        )

    def wrap_dataset(self, lit_module, dataset):
        """Wrap dataset for baseline evaluation"""
        return self.baseline.wrap_dataset(
            dataset,
            self.env,
            batch_size=lit_module.val_batch_size,
            device=get_lightning_device(lit_module),
        )



import torch.nn as nn

from tensordict import TensorDict
from torchrl.envs import EnvBase

from rl4co.models.nn.env_embeddings import env_init_embedding
from rl4co.models.zoo.mdam.decoder import Decoder
from rl4co.models.zoo.mdam.encoder import GraphAttentionEncoder


class MDAMPolicy(nn.Module):
    """
    Args:
        env: environment to solve
        encoder: encoder module
        decoder: decoder module
        embedding_dim: embedding dimension/hidden dimension
        num_encode_layers: number of layers in encoder
        num_heads: number of heads in multi-head attention
        num_paths: number of paths to sample (specific feature for MDAM)
        eg_step_gap: number of steps between each path sampling (specific feature for MDAM)
        normalization: normalization type
        mask_inner: whether to mask the inner product in attention
        mask_logits: whether to mask the logits in attention
        tanh_clipping: tanh clipping value
        shrink_size: shrink size for the decoder
        use_native_sdpa: whether to use native sdpa (scaled dot product attention)
        force_flash_attn: whether to force use flash attention
        train_decode_type: decode type for training
        val_decode_type: decode type for validation
        test_decode_type: decode type for testing
    """

    def __init__(
        self,
        env: EnvBase,
        encoder: nn.Module = None,
        decoder: nn.Module = None,
        embedding_dim: int = 128,
        num_encode_layers: int = 3,
        num_heads: int = 8,
        num_paths: int = 5,
        eg_step_gap: int = 200,
        normalization: str = "batch",
        mask_inner: bool = True,
        mask_logits: bool = True,
        tanh_clipping: float = 10.0,
        shrink_size=None,
        use_native_sdpa: bool = False,
        force_flash_attn: bool = False,
        train_decode_type: str = "sampling",
        val_decode_type: str = "greedy",
        test_decode_type: str = "greedy",
        **unused_kw,
    ):
        super(MDAMPolicy, self).__init__()
        if len(unused_kw) > 0:
            print(f"Unused kwargs: {unused_kw}")

        self.env = env
        self.init_embedding = env_init_embedding(
            self.env.name, {"embedding_dim": embedding_dim}
        )

        self.encoder = (
            GraphAttentionEncoder(
                num_heads=num_heads,
                embed_dim=embedding_dim,
                num_layers=num_encode_layers,
                normalization=normalization,
                use_native_sdpa=use_native_sdpa,
                force_flash_attn=force_flash_attn,
            )
            if encoder is None
            else encoder
        )

        self.decoder = (
            Decoder(
                env=env,
                embedding_dim=embedding_dim,
                num_heads=num_heads,
                num_paths=num_paths,
                mask_inner=mask_inner,
                mask_logits=mask_logits,
                eg_step_gap=eg_step_gap,
                tanh_clipping=tanh_clipping,
                force_flash_attn=force_flash_attn,
                shrink_size=shrink_size,
                train_decode_type=train_decode_type,
                val_decode_type=val_decode_type,
                test_decode_type=test_decode_type,
            )
            if decoder is None
            else decoder
        )

    def forward(
        self,
        td: TensorDict,
        phase: str = "train",
        return_actions: bool = False,
        **decoder_kwargs,
    ) -> TensorDict:
        embedding = self.init_embedding(td)
        encoded_inputs, _, attn, V, h_old = self.encoder(embedding)

        # Get decode type depending on phase
        if decoder_kwargs.get("decode_type", None) is None:
            decoder_kwargs["decode_type"] = getattr(self, f"{phase}_decode_type")

        reward, log_likelihood, kl_divergence, actions = self.decoder(
            td, encoded_inputs, attn, V, h_old, **decoder_kwargs
        )
        out = {
            "reward": reward,
            "log_likelihood": log_likelihood,
            "kl_divergence": kl_divergence,
            "actions": actions if return_actions else None,
        }
        return out



import os
import zipfile

from typing import Optional

import numpy as np
import torch

from tensordict.tensordict import TensorDict
from torchrl.data import (
    BoundedTensorSpec,
    CompositeSpec,
    UnboundedContinuousTensorSpec,
    UnboundedDiscreteTensorSpec,
)

from rl4co.data.utils import load_npz_to_tensordict
from rl4co.envs.common.base import RL4COEnvBase
from rl4co.utils.download.downloader import download_url
from rl4co.utils.pylogger import get_pylogger

log = get_pylogger(__name__)


class DPPEnv(RL4COEnvBase):
    """Decap placement problem as done in DevFormer paper

    https://arxiv.org/abs/2205.13225
    """

    name = "dpp"

    def __init__(
        self,
        *,
        min_loc: float = 0,
        max_loc: float = 1,
        num_keepout_min: int = 1,
        num_keepout_max: int = 50,
        max_decaps: int = 20,
        data_dir: str = "data/dpp/",
        chip_file: str = "10x10_pkg_chip.npy",
        decap_file: str = "01nF_decap.npy",
        freq_file: str = "freq_201.npy",
        url: str = None,
        td_params: TensorDict = None,
        **kwargs,
    ):
        kwargs["data_dir"] = data_dir
        super().__init__(**kwargs)

        self.url = (
            "https://drive.google.com/uc?id=1IEuR2v8Le-mtHWHxwTAbTOPIkkQszI95"
            if url is None
            else url
        )
        self._load_dpp_data(chip_file, decap_file, freq_file)
        self.min_loc = min_loc
        self.max_loc = max_loc
        self.num_keepout_min = num_keepout_min
        self.num_keepout_max = num_keepout_max
        self.max_decaps = max_decaps

        assert (
            num_keepout_min <= num_keepout_max
        ), "num_keepout_min must be <= num_keepout_max"
        assert (
            num_keepout_max <= self.size**2
        ), "num_keepout_max must be <= size * size (total number of locations)"

        self._make_spec(td_params)

    def _step(self, td: TensorDict) -> TensorDict:
        current_node = td["action"]

        # Set available to 0 (i.e., already placed) if the current node is the first node
        available = td["action_mask"].scatter(
            -1, current_node.unsqueeze(-1).expand_as(td["action_mask"]), 0
        )

        # Set done if i is greater than max_decaps
        done = td["i"] >= self.max_decaps - 1

        # Calculate reward (we set to -inf since we calculate the reward outside based on the actions)
        reward = torch.ones_like(done) * float("-inf")

        # The output must be written in a ``"next"`` entry
        return TensorDict(
            {
                "next": {
                    "locs": td["locs"],
                    "probe": td["probe"],
                    "i": td["i"] + 1,
                    "action_mask": available,
                    "keepout": td["keepout"],
                    "reward": reward,
                    "done": done,
                }
            },
            td.shape,
        )

    def _reset(self, td: Optional[TensorDict] = None, batch_size=None) -> TensorDict:
        # Initialize locations
        if batch_size is None:
            batch_size = self.batch_size if td is None else td.batch_size
        self.device = td.device if td is not None else self.device

        # We allow loading the initial observation from a dataset for faster loading
        if td is None:
            td = self.generate_data(batch_size=batch_size)

        # Other variables
        i = torch.zeros((*batch_size, 1), dtype=torch.int64, device=self.device)

        return TensorDict(
            {
                "locs": td["locs"],
                "probe": td["probe"],
                "i": i,
                "action_mask": td["action_mask"],
                "keepout": ~td["action_mask"],
            },
            batch_size=batch_size,
        )

    def _make_spec(self, td_params):
        """Make the observation and action specs from the parameters"""
        self.observation_spec = CompositeSpec(
            locs=BoundedTensorSpec(
                minimum=self.min_loc,
                maximum=self.max_loc,
                shape=(self.size**2, 2),
                dtype=torch.float32,
            ),
            probe=UnboundedDiscreteTensorSpec(
                shape=(1),
                dtype=torch.int64,
            ),
            keepout=UnboundedDiscreteTensorSpec(
                shape=(self.size**2),
                dtype=torch.bool,
            ),
            i=UnboundedDiscreteTensorSpec(
                shape=(1),
                dtype=torch.int64,
            ),
            action_mask=UnboundedDiscreteTensorSpec(
                shape=(self.size**2),
                dtype=torch.bool,
            ),
            shape=(),
        )
        self.input_spec = self.observation_spec.clone()
        self.action_spec = BoundedTensorSpec(
            shape=(1,),
            dtype=torch.int64,
            minimum=0,
            maximum=self.size**2,
        )
        self.reward_spec = UnboundedContinuousTensorSpec(shape=(1,))
        self.done_spec = UnboundedDiscreteTensorSpec(shape=(1,), dtype=torch.bool)

    def get_reward(self, td, actions):
        """
        We call the reward function with the final sequence of actions to get the reward
        Calling per-step would be very time consuming due to decap simulation
        """
        # We do the operation in a batch
        if len(td.batch_size) == 0:
            td = td.unsqueeze(0)
            actions = actions.unsqueeze(0)
        probes = td["probe"]
        reward = torch.stack(
            [self._decap_simulator(p, a) for p, a in zip(probes, actions)]
        )
        return reward

    def generate_data(self, batch_size):
        """
        Generate initial observations for the environment with locations, probe, and action mask
        Action_mask eliminates the keepout regions and the probe location, and is updated to eliminate placed decaps
        """
        m = n = self.size
        # if int, convert to list and make it a batch for easier generation
        batch_size = [batch_size] if isinstance(batch_size, int) else batch_size
        batched = len(batch_size) > 0
        bs = [1] if not batched else batch_size

        # Create a list of locs on a grid
        locs = torch.meshgrid(
            torch.arange(m, device=self.device), torch.arange(n, device=self.device)
        )
        locs = torch.stack(locs, dim=-1).reshape(-1, 2)
        # normalize the locations by the number of rows and columns
        locs = locs / torch.tensor([m, n], dtype=torch.float, device=self.device)
        locs = locs[None].expand(*bs, -1, -1)

        # Create available mask
        available = torch.ones((*bs, m * n), dtype=torch.bool)

        # Sample probe location from m*n
        probe = torch.randint(m * n, size=(*bs, 1))
        available.scatter_(1, probe, False)

        # Sample keepout locations from m*n except probe
        num_keepout = torch.randint(
            self.num_keepout_min,
            self.num_keepout_max,
            size=(*bs, 1),
            device=self.device,
        )
        keepouts = [torch.randperm(m * n)[:k] for k in num_keepout]
        for i, (a, k) in enumerate(zip(available, keepouts)):
            available[i] = a.scatter(0, k, False)

        return TensorDict(
            {
                "locs": locs if batched else locs.squeeze(0),
                "probe": probe if batched else probe.squeeze(0),
                "action_mask": available if batched else available.squeeze(0),
            },
            batch_size=batch_size,
        )

    def _decap_placement(self, pi, probe):
        device = pi.device

        n = m = self.size  # columns and rows
        num_decap = torch.numel(pi)
        z1 = self.raw_pdn.to(device)

        decap = self.decap.reshape(-1).to(device)
        z2 = torch.zeros(
            (self.num_freq, num_decap, num_decap), dtype=torch.float32, device=device
        )

        qIndx = torch.arange(num_decap, device=device)

        z2[:, qIndx, qIndx] = torch.abs(decap)[:, None].repeat_interleave(
            z2[:, qIndx, qIndx].shape[-1], dim=-1
        )
        pIndx = pi.long()

        aIndx = torch.arange(len(z1[0]), device=device)
        aIndx = torch.tensor(
            list(set(aIndx.tolist()) - set(pIndx.tolist())), device=device
        )

        z1aa = z1[:, aIndx, :][:, :, aIndx]
        z1ap = z1[:, aIndx, :][:, :, pIndx]
        z1pa = z1[:, pIndx, :][:, :, aIndx]
        z1pp = z1[:, pIndx, :][:, :, pIndx]
        z2qq = z2[:, qIndx, :][:, :, qIndx]

        zout = z1aa - torch.matmul(torch.matmul(z1ap, torch.inverse(z1pp + z2qq)), z1pa)

        idx = torch.arange(n * m, device=device)
        mask = torch.zeros(n * m, device=device).bool()
        mask[pi] = True
        mask = mask & (idx < probe)
        probe -= mask.sum().item()

        zout = zout[:, probe, probe]
        return zout

    def _decap_model(self, z_initial, z_final):
        impedance_gap = torch.zeros(self.num_freq, device=self.device)

        impedance_gap = z_initial - z_final
        reward = torch.sum(impedance_gap * 1000000000 / self.freq.to(self.device))

        reward = reward / 10
        return reward

    def _initial_impedance(self, probe):
        zout = self.raw_pdn.to(self.device)[:, probe, probe]
        return zout

    def _decap_simulator(self, probe, solution, keepout=None):
        self.device = solution.device

        probe = probe.item()

        assert len(solution) == len(
            torch.unique(solution)
        ), "An Element of Decap Sequence must be Unique"

        if keepout is not None:
            keepout = torch.tensor(keepout)
            intersect = torch.tensor(list(set(solution.tolist()) & set(keepout.tolist())))
            assert len(intersect) == 0, "Decap must be not placed at the keepout region"

        z_initial = self._initial_impedance(probe)
        z_initial = torch.abs(z_initial)
        z_final = self._decap_placement(solution, probe)
        z_final = torch.abs(z_final)
        reward = self._decap_model(z_initial, z_final)
        return reward

    def _load_dpp_data(self, chip_file, decap_file, freq_file):
        def _load_file(fpath):
            f = os.path.join(self.data_dir, fpath)
            if not os.path.isfile(f):
                self._download_data()
            with open(f, "rb") as f_:
                return torch.from_numpy(np.load(f_)).to(self.device)

        self.raw_pdn = _load_file(chip_file)  # [num_freq, size^2, size^2]
        self.decap = _load_file(decap_file).to(torch.complex64)  # [num_freq, 1, 1]
        self.freq = _load_file(freq_file)  # [num_freq]
        self.size = int(np.sqrt(self.raw_pdn.shape[-1]))
        self.num_freq = self.freq.shape[0]

    def _download_data(self):
        log.info("Downloading data...")
        download_url(self.url, self.data_dir, "data.zip")
        log.info("Download complete. Unzipping...")
        zipfile.ZipFile(os.path.join(self.data_dir, "data.zip"), "r").extractall(
            self.data_dir
        )
        log.info("Unzip complete. Removing zip file")
        os.remove(os.path.join(self.data_dir, "data.zip"))

    def load_data(self, fpath, batch_size=[]):
        data = load_npz_to_tensordict(fpath)
        # rename key if necessary (old dpp version)
        if "observation" in data.keys():
            data["locs"] = data.pop("observation")
        return data

    def render(self, decaps, probe, action_mask, ax=None, legend=True):
        """
        Plot a grid of 1x1 squares representing the environment.
        The keepout regions are the action_mask - decaps - probe
        """
        import matplotlib.pyplot as plt

        settings = {
            0: {"color": "white", "label": "available"},
            1: {"color": "grey", "label": "keepout"},
            2: {"color": "tab:red", "label": "probe"},
            3: {"color": "tab:blue", "label": "decap"},
        }

        nonzero_indices = torch.nonzero(~action_mask, as_tuple=True)[0]
        keepout = torch.cat([nonzero_indices, probe, decaps.squeeze(-1)])
        unique_elements, counts = torch.unique(keepout, return_counts=True)
        keepout = unique_elements[counts == 1]

        if ax is None:
            fig, ax = plt.subplots(1, 1, figsize=(6, 6))

        grid = np.meshgrid(np.arange(0, self.size), np.arange(0, self.size))
        grid = np.stack(grid, axis=-1)

        # Add new dimension to grid filled up with 0s
        grid = np.concatenate([grid, np.zeros((self.size, self.size, 1))], axis=-1)

        # Add keepout = 1
        grid[keepout // self.size, keepout % self.size, 2] = 1
        # Add probe = 2
        grid[probe // self.size, probe % self.size, 2] = 2
        # Add decaps = 3
        grid[decaps // self.size, decaps % self.size, 2] = 3

        xdim, ydim = grid.shape[0], grid.shape[1]
        ax.imshow(np.zeros((xdim, ydim)), cmap="gray")

        ax.set_xlim(0, xdim)
        ax.set_ylim(0, ydim)

        for i in range(xdim):
            for j in range(ydim):
                color = settings[grid[i, j, 2]]["color"]
                x, y = grid[i, j, 0], grid[i, j, 1]
                ax.add_patch(plt.Rectangle((x, y), 1, 1, color=color, linestyle="-"))

        # Add grid with 1x1 squares
        ax.grid(
            which="major", axis="both", linestyle="-", color="k", linewidth=1, alpha=0.5
        )
        # set 10 ticks
        ax.set_xticks(np.arange(0, xdim, 1))
        ax.set_yticks(np.arange(0, ydim, 1))

        # Invert y axis
        ax.invert_yaxis()

        # Add legend
        if legend:
            num_unique = 4
            handles = [
                plt.Rectangle((0, 0), 1, 1, color=settings[i]["color"])
                for i in range(num_unique)
            ]
            ax.legend(
                handles,
                [settings[i]["label"] for i in range(num_unique)],
                ncol=num_unique,
                loc="upper center",
                bbox_to_anchor=(0.5, 1.1),
            )



import math

import torch
import torch.nn as nn

from tensordict.tensordict import TensorDict

from rl4co.utils.ops import batchify


def rotation_reflection_transform(x, y, phi, offset=0.5):
    """SR group transform with rotation and reflection (~2x faster than original)"""
    x, y = x - offset, y - offset
    # random rotation
    x_prime = torch.cos(phi) * x - torch.sin(phi) * y
    y_prime = torch.sin(phi) * x + torch.cos(phi) * y
    # make random reflection if phi > 2*pi (i.e. 50% of the time)
    mask = phi > 2 * math.pi
    # vectorized random reflection: swap axes x and y if mask
    xy = torch.cat((x_prime, y_prime), dim=-1)
    xy = torch.where(mask, xy.flip(-1), xy)
    return xy + offset


def augment_xy_data_by_n_fold(xy, num_augment: int = 8):
    """Augment xy data by N times via symmetric rotation transform and concatenate to original data"""
    # create random rotation angles (4*pi for reflection, 2*pi for rotation)
    phi = torch.rand(xy.shape[0], device=xy.device) * 4 * math.pi
    # set phi to 0 for first , i.e. no augmnetation as in original paper
    phi[: xy.shape[0] // num_augment] = 0.0
    x, y = xy[..., [0]], xy[..., [1]]
    return rotation_reflection_transform(x, y, phi[:, None, None])


def env_aug_feats(env_name: str):
    return ("locs", "depot") if env_name == "op" else ("locs",)


def min_max_normalize(x):
    return (x - x.min()) / (x.max() - x.min())


class StateAugmentation(nn.Module):
    """Augment state by N times via symmetric rotation/reflection transform"""

    def __init__(self, env_name: str, num_augment: int = 8, normalize: bool = False):
        super(StateAugmentation, self).__init__()
        self.augmentation = augment_xy_data_by_n_fold
        self.feats = env_aug_feats(env_name)
        self.num_augment = num_augment
        self.normalize = normalize

    def forward(
        self, td: TensorDict, num_augment: int = None, normalize: bool = False
    ) -> TensorDict:
        num_augment = num_augment if num_augment is not None else self.num_augment
        normalize = normalize if normalize is not None else False

        td_aug = batchify(td, num_augment)
        for feat in self.feats:
            aug_feat = self.augmentation(td_aug[feat], num_augment)
            td_aug[feat] = aug_feat
            if normalize:
                td_aug[feat] = min_max_normalize(td_aug[feat])

        return td_aug



from math import log
from typing import Union

import torch
import torch.nn as nn
import torch.nn.functional as F

from tensordict import TensorDict

from rl4co.utils.pylogger import get_pylogger

log = get_pylogger(__name__)


class PPO(nn.Module):
    def __init__(
        self,
        env,
        policy: nn.Module,
        critic: nn.Module,
        clip_range: float = 0.2,  # epsilon of PPO
        ppo_epochs: int = 2,  # K
        mini_batch_size: Union[int, float] = 0.25,  # 0.25,
        vf_lambda: float = 0.5,  # lambda of Value function fitting
        entropy_lambda: float = 0.0,  # lambda of entropy bonus
        normalize_adv: bool = False,  # whether to normalize advantage
        max_grad_norm: float = 0.5,  # max gradient norm
        **unused_kw,
    ):
        super().__init__()
        if len(unused_kw) > 0:
            log.warn(f"Unused kwargs: {unused_kw}")
        self.env = env
        self.policy = policy
        self.critic = critic

        # PPO hyper params
        self.clip_range = clip_range
        self.ppo_epochs = ppo_epochs
        self.mini_batch_size = mini_batch_size
        self.vf_lambda = vf_lambda
        self.entropy_lambda = entropy_lambda
        self.normalize_adv = normalize_adv
        self.max_grad_norm = max_grad_norm

    def forward(
        self,
        td: TensorDict,
        phase: str = "train",
        extra=None,
        policy_kwargs: dict = {},
        critic_kwargs: dict = {},
        optimizer=None,
    ):
        # Evaluate model, get costs and log probabilities
        with torch.no_grad():
            # compute a_old and logp_old
            out = self.policy(td.clone(), phase, return_action=True, **policy_kwargs)
            old_logp = out["log_likelihood"]  # [batch, decoder steps]
            actions = out["actions"]  # [batch, decoder steps]
            rewards = out["reward"]  # [batch]

        iter_i = 0
        if phase == "train":
            batch_size = old_logp.shape[0]

            if isinstance(self.mini_batch_size, float):
                mini_batch_size = int(self.mini_batch_size * batch_size)
            if self.mini_batch_size >= batch_size:
                mini_batch_size = batch_size

            for _ in range(self.ppo_epochs):  # loop K
                for mini_batch_idx in torch.randperm(batch_size).split(mini_batch_size):
                    # compute a and logp
                    mini_batched_out = self.policy(
                        td[mini_batch_idx].clone(),
                        phase,
                        given_actions=actions[mini_batch_idx],
                        return_entropy=True,
                        calc_reward=False,
                        **policy_kwargs,
                    )

                    # compute ratio
                    ratio = torch.exp(
                        mini_batched_out["selected_log_p"].sum(dim=-1)
                        - old_logp[mini_batch_idx].sum(dim=-1)
                    )  # [batch size]

                    # compute advantage

                    value_pred = self.critic(td[mini_batch_idx], **critic_kwargs)
                    adv = rewards[mini_batch_idx] - value_pred.detach()  # [batch size]

                    if self.normalize_adv:
                        adv = (adv - adv.mean()) / (adv.std() + 1e-6)

                    # compute surrogate loss
                    surrogate_loss = -torch.min(
                        ratio * adv,
                        torch.clamp(ratio, 1 - self.clip_range, 1 + self.clip_range)
                        * adv,
                    ).mean()

                    # compute entropy bonus
                    entropy_bonus = mini_batched_out["entropy"].mean()

                    # compute value function loss
                    value_loss = F.huber_loss(
                        value_pred, rewards[mini_batch_idx].view(-1, 1)
                    )

                    # compute total loss
                    loss = (
                        surrogate_loss
                        + self.vf_lambda * value_loss
                        - self.entropy_lambda * entropy_bonus
                    )

                    # perform optimization
                    if optimizer is not None:
                        optimizer.zero_grad()
                        loss.backward()
                        if self.max_grad_norm is not None:
                            nn.utils.clip_grad_norm_(
                                self.parameters(), self.max_grad_norm
                            )
                        optimizer.step()

                        iter_i += 1

            # log training results
            out.update(
                {
                    "loss": loss,
                    "surrogate_loss": surrogate_loss,
                    "value_loss": value_loss,
                    "entropy_bonus": entropy_bonus,
                }
            )
        return out



from rl4co.models.rl.ppo.model import PPO
from rl4co.models.rl.reinforce.critic import CriticNetwork
from rl4co.models.zoo.amppo.policy import PPOAttentionModelPolicy


class AttentionModel(PPO):
    def __init__(self, env, policy=None, critic=None, **policy_kwargs):
        policy = (
            PPOAttentionModelPolicy(env=env, **policy_kwargs)
            if policy is None
            else policy
        )
        critic = CriticNetwork(env=env) if critic is None else critic
        super(AttentionModel, self).__init__(
            env=env,
            policy=policy,
            critic=critic,
            **policy_kwargs,
        )



import copy

import torch
import torch.nn as nn
import torch.nn.functional as F

from scipy.stats import ttest_rel
from torch.utils.data import DataLoader
from tqdm.auto import tqdm

from rl4co import utils
from rl4co.data.dataset import ExtraKeyDataset, tensordict_collate_fn

log = utils.get_pylogger(__name__)


class REINFORCEBaseline(nn.Module):
    """Base class for REINFORCE baselines"""

    def __init__(self, *args, **kw):
        super().__init__()
        pass

    def wrap_dataset(self, dataset, *args, **kw):
        """Wrap dataset with baseline-specific functionality"""
        return dataset

    def eval(self, td, reward):
        """Evaluate baseline"""
        pass

    def epoch_callback(self, *args, **kw):
        """Callback at the end of each epoch
        For example, update baseline parameters and obtain baseline values
        """
        pass

    def setup(self, *args, **kw):
        """To be called before training during setup phase
        This follow PyTorch Lightning's setup() convention
        """
        pass


class NoBaseline(REINFORCEBaseline):
    def eval(self, td, reward):
        return 0, 0  # No baseline, no neg_los


class SharedBaseline(REINFORCEBaseline):
    def eval(self, td, reward, on_dim=1):  # e.g. [batch, pomo, ...]
        return reward.mean(dim=on_dim, keepdims=True), 0


class ExponentialBaseline(REINFORCEBaseline):
    def __init__(self, beta=0.8):
        super(REINFORCEBaseline, self).__init__()

        self.beta = beta
        self.v = None

    def eval(self, td, reward):
        if self.v is None:
            v = reward.mean()
        else:
            v = self.beta * self.v + (1.0 - self.beta) * reward.mean()
        self.v = v.detach()  # Detach since we never want to backprop
        return self.v, 0  # No loss


class WarmupBaseline(REINFORCEBaseline):
    def __init__(
        self,
        baseline,
        n_epochs=1,
        warmup_exp_beta=0.8,
    ):
        super(REINFORCEBaseline, self).__init__()

        self.baseline = baseline
        assert n_epochs > 0, "n_epochs to warmup must be positive"
        self.warmup_baseline = ExponentialBaseline(warmup_exp_beta)
        self.alpha = 0
        self.n_epochs = n_epochs

    def wrap_dataset(self, dataset, *args, **kw):
        if self.alpha > 0:
            return self.baseline.wrap_dataset(dataset, *args, **kw)
        return self.warmup_baseline.wrap_dataset(dataset, *args, **kw)

    def setup(self, *args, **kw):
        self.baseline.setup(*args, **kw)

    def eval(self, td, reward):
        if self.alpha == 1:
            return self.baseline.eval(td, reward)
        if self.alpha == 0:
            return self.warmup_baseline.eval(td, reward)
        v_b, l_b = self.baseline.eval(td, reward)
        v_wb, l_wb = self.warmup_baseline.eval(td, reward)
        # Return convex combination of baseline and of loss
        return self.alpha * v_b + (1 - self.alpha) * v_wb, self.alpha * l_b + (
            1 - self.alpha * l_wb
        )

    def epoch_callback(self, *args, **kw):
        # Need to call epoch callback of inner model (also after first epoch if we have not used it)
        self.baseline.epoch_callback(*args, **kw)
        self.alpha = (kw["epoch"] + 1) / float(self.n_epochs)
        if kw["epoch"] < self.n_epochs:
            log.info("Set warmup alpha = {}".format(self.alpha))


class CriticBaseline(REINFORCEBaseline):
    def __init__(self, critic, **unused_kw):
        super(CriticBaseline, self).__init__()
        self.critic = critic

    def eval(self, x, c):
        v = self.critic(x)
        # detach v since actor should not backprop through baseline, only for neg_loss
        return v.detach(), -F.mse_loss(v, c.detach())


class RolloutBaseline(REINFORCEBaseline):
    def __init__(self, bl_alpha=0.05, progress_bar=False):
        super(RolloutBaseline, self).__init__()
        self.bl_alpha = bl_alpha
        self.progress_bar = progress_bar

    def setup(self, *args, **kw):
        self._update_model(*args, **kw)

    def _update_model(
        self, model, env, batch_size=64, device="cpu", dataset_size=None, dataset=None
    ):
        self.model = copy.deepcopy(model).to(device)
        if dataset is None:
            log.info("Creating evaluation dataset for rollout baseline")
            self.dataset = env.dataset(batch_size=[dataset_size])

        log.info("Evaluating baseline model on evaluation dataset")
        self.bl_vals = (
            self.rollout(self.model, env, batch_size, device, self.dataset).cpu().numpy()
        )
        self.mean = self.bl_vals.mean()

    def eval(self, td, reward):
        # Use volatile mode for efficient inference (single batch so we do not use rollout function)
        with torch.no_grad():
            reward = self.model(td)["reward"]
        return reward, 0

    def epoch_callback(
        self, model, env, batch_size=64, device="cpu", epoch=None, dataset_size=None
    ):
        """Challenges the current baseline with the model and replaces the baseline model if it is improved"""
        log.info("Evaluating candidate model on evaluation dataset")
        candidate_vals = self.rollout(model, env, batch_size, device).cpu().numpy()
        candidate_mean = candidate_vals.mean()

        log.info(
            "Candidate mean: {:.3f}, Baseline mean: {:.3f}".format(
                candidate_mean, self.mean
            )
        )
        if candidate_mean - self.mean > 0:
            # Calc p value with inverse logic (costs)
            t, p = ttest_rel(-candidate_vals, -self.bl_vals)

            p_val = p / 2  # one-sided
            assert t < 0, "T-statistic should be negative"
            log.info("p-value: {:.3f}".format(p_val))
            if p_val < self.bl_alpha:
                log.info("Updating baseline")
                self._update_model(model, env, batch_size, device, dataset_size)

    def rollout(self, model, env=None, batch_size=64, device="cpu", dataset=None):
        """Rollout the model on the given dataset"""
        # if dataset is None, use the dataset of the baseline
        dataset = self.dataset if dataset is None else dataset

        model.eval()
        model = model.to(device)

        def eval_model(batch):
            with torch.no_grad():
                batch = env.reset(batch.to(device))
                return model(batch, decode_type="greedy")["reward"].data.cpu()

        dl = DataLoader(dataset, batch_size=batch_size, collate_fn=tensordict_collate_fn)

        retval = torch.cat(
            [eval_model(batch) for batch in tqdm(dl, disable=not self.progress_bar)], 0
        )
        return retval

    def wrap_dataset(self, dataset, env, batch_size=64, device="cpu", **kw):
        """Wrap the dataset in a baseline dataset"""
        rewards = (
            self.rollout(self.model, env, batch_size, device, dataset=dataset)
            .detach()
            .cpu()
        )
        return ExtraKeyDataset(dataset, rewards)

    def __getstate__(self):
        """Do not include datasets in state to avoid pickling issues"""
        state = self.__dict__.copy()
        try:
            del state["dataset"]
        except KeyError:
            pass
        return state

    def __setstate__(self, state):
        """Restore datasets after unpickling. Will be restored in setup"""
        self.__dict__.update(state)
        self.dataset = None



from typing import Any

import torch.nn as nn

from omegaconf import DictConfig

from rl4co.envs.base import EnvBase
from rl4co.tasks.rl4co import RL4COLitModule


class PPOTask(RL4COLitModule):
    def __init__(self, cfg: DictConfig, env: EnvBase = None, model: nn.Module = None):
        super().__init__(cfg=cfg, env=env, model=model)
        self.automatic_optimization = False

    def shared_step(self, batch: Any, batch_idx: int, phase: str):
        td = self.env.reset(batch)
        out = self.model(
            td,
            phase,
            td.get("extra", None),
            optimizer=self.optimizers() if phase == "train" else None,
        )

        # Log metrics
        metrics = getattr(self, f"{phase}_metrics")
        metrics = {f"{phase}/{k}": v.mean() for k, v in out.items() if k in metrics}

        log_on_step = self.log_on_step if phase == "train" else False
        on_epoch = False if phase == "train" else True
        self.log_dict(
            metrics,
            on_step=log_on_step,
            on_epoch=on_epoch,
            prog_bar=True,
            sync_dist=True,
            add_dataloader_idx=False,
        )
        return {"loss": out.get("loss", None), **metrics}



import torch

from rl4co.models.nn.utils import decode_probs
from rl4co.models.zoo.am.decoder import Decoder


class PPODecoder(Decoder):

    """
    A slightly modified AM decoder to support PPO training.
    """

    def forward(
        self,
        td,
        embeddings,
        decode_type="sampling",
        softmax_temp=None,
        calc_reward: bool = True,
        given_actions: torch.Tensor = None,  # [batch_size, graph_size]
    ):
        outputs = []
        actions = []

        # Compute keys, values for the glimpse and keys for the logits once as they can be reused in every step
        cached_embeds = self._precompute(embeddings)

        decode_step = 0
        while not td["done"].all():
            log_p, mask = self._get_log_p(cached_embeds, td, softmax_temp)

            # Select the indices of the next nodes in the sequences, result (batch_size) long

            if given_actions is not None:
                action = given_actions[..., decode_step]
            else:
                action = decode_probs(log_p.exp(), mask, decode_type=decode_type)

            td.set("action", action)
            td = self.env.step(td)["next"]

            outputs.append(log_p)
            actions.append(action)

            decode_step += 1

        if given_actions is not None:
            if len(outputs) != given_actions.shape[1]:
                # print(given_actions.shape, decode_step)
                # print(td["done"].all())
                raise ValueError(
                    f"Given actions have {given_actions.shape[1]} steps, but we decoded {decode_step} steps."
                )

        # output: logprobs [batch, problem size, decoding steps]
        outputs, actions = torch.stack(outputs, 1), torch.stack(actions, 1)
        if calc_reward:
            td.set("reward", self.env.get_reward(td, actions))

        return outputs, actions, td



import pytest

from rl4co.envs import ATSPEnv, CVRPEnv, DPPEnv, MTSPEnv, PDPEnv, SDVRPEnv, TSPEnv
from rl4co.models.nn.utils import random_policy, rollout


@pytest.mark.parametrize("size, batch_size", [(20, 2)])
def test_tsp(size, batch_size):
    env = TSPEnv(num_loc=size)
    reward = rollout(env, env.reset(batch_size=[batch_size]), random_policy)
    assert reward.shape == (batch_size,)


@pytest.mark.parametrize("size, batch_size", [(20, 2)])
def test_atsp(size, batch_size):
    env = ATSPEnv(num_loc=size)
    reward = rollout(env, env.reset(batch_size=[batch_size]), random_policy)
    assert reward.shape == (batch_size,)


@pytest.mark.parametrize("size, batch_size", [(20, 2)])
def test_dpp(size, batch_size):
    env = DPPEnv()
    reward = rollout(env, env.reset(batch_size=[batch_size]), random_policy)
    assert reward.shape == (batch_size,)


@pytest.mark.parametrize("size, batch_size", [(20, 2)])
def test_cvrp(size, batch_size):
    env = CVRPEnv()
    reward = rollout(env, env.reset(batch_size=[batch_size]), random_policy)
    assert reward.shape == (batch_size,)


@pytest.mark.parametrize("size, batch_size", [(20, 2)])
def test_sdvrp(size, batch_size):
    env = SDVRPEnv()
    reward = rollout(env, env.reset(batch_size=[batch_size]), random_policy)
    assert reward.shape == (batch_size,)


@pytest.mark.parametrize("size, batch_size", [(20, 2)])
def test_pdp(size, batch_size):
    env = PDPEnv(num_loc=size)
    reward = rollout(env, env.reset(batch_size=[batch_size]), random_policy)
    assert reward.shape == (batch_size,)


@pytest.mark.parametrize("size, batch_size", [(20, 2)])
def test_mtsp(size, batch_size):
    env = MTSPEnv(num_loc=size)
    reward = rollout(env, env.reset(batch_size=[batch_size]), random_policy)
    assert reward.shape == (batch_size,)



import torch.nn as nn

from rl4co.models.nn.attention import MultiHeadAttention, NativeFlashMHA
from rl4co.models.nn.env_embeddings import env_init_embedding
from rl4co.models.nn.ops import Normalization, SkipConnection
from rl4co.utils.pylogger import get_pylogger

log = get_pylogger(__name__)


class MultiHeadAttentionLayer(nn.Sequential):
    def __init__(
        self,
        num_heads,
        embed_dim,
        feed_forward_hidden=512,
        normalization="batch",
        use_native_sdpa=False,
        force_flash_attn=False,
    ):
        MHA = NativeFlashMHA if use_native_sdpa else MultiHeadAttention
        super(MultiHeadAttentionLayer, self).__init__(
            SkipConnection(MHA(embed_dim, num_heads, force_flash_attn=force_flash_attn)),
            Normalization(embed_dim, normalization),
            SkipConnection(
                nn.Sequential(
                    nn.Linear(embed_dim, feed_forward_hidden),
                    nn.ReLU(),
                    nn.Linear(feed_forward_hidden, embed_dim),
                )
                if feed_forward_hidden > 0
                else nn.Linear(embed_dim, embed_dim)
            ),
            Normalization(embed_dim, normalization),
        )


class GraphAttentionEncoder(nn.Module):
    """Graph Attention Encoder with a series of MHA layers
    Multi-Head Attention Layer with normalization and feed-forward layer
    If use_native_sdpa is True, use NativeFlashMHA instead of MultiHeadAttention:
    native PyTorch `scaled_dot_product_attention` implementation, available from 2.0
    You may force FlashAttention by setting force_flash_attn to True (move to half precision)
    """

    def __init__(
        self,
        num_heads,
        embedding_dim,
        num_layers,
        env=None,
        normalization="batch",
        feed_forward_hidden=512,
        use_native_sdpa=False,
        force_flash_attn=False,
        disable_init_embedding=False,
    ):
        super(GraphAttentionEncoder, self).__init__()

        # To map input to embedding space
        if not disable_init_embedding:
            self.init_embedding = env_init_embedding(
                env.name, {"embedding_dim": embedding_dim}
            )
        else:
            log.warning("Disabling init embedding manually for GraphAttentionEncoder")
            self.init_embedding = nn.Identity()  # do nothing

        self.layers = nn.Sequential(
            *(
                MultiHeadAttentionLayer(
                    num_heads,
                    embedding_dim,
                    feed_forward_hidden=feed_forward_hidden,
                    normalization=normalization,
                    use_native_sdpa=use_native_sdpa,
                    force_flash_attn=force_flash_attn,
                )
                for _ in range(num_layers)
            )
        )

    def forward(self, x, mask=None):
        assert mask is None, "Mask not yet supported!"
        # initial Embedding from features
        init_embeds = self.init_embedding(x)
        # layers  (batch_size, graph_size, embed_dim)
        embeds = self.layers(init_embeds)
        return embeds, init_embeds



from tensordict import TensorDict

from rl4co.models.rl.reinforce.base import REINFORCE
from rl4co.models.rl.reinforce.baselines import NoBaseline
from rl4co.models.zoo.symnco.augmentations import StateAugmentation
from rl4co.models.zoo.symnco.losses import (
    invariance_loss,
    problem_symmetricity_loss,
    solution_symmetricity_loss,
)
from rl4co.models.zoo.symnco.policy import SymNCOPolicy
from rl4co.utils.ops import gather_by_index, unbatchify
from rl4co.utils.pylogger import get_pylogger

log = get_pylogger(__name__)


class SymNCO(REINFORCE):
    """SymNCO Model for neural combinatorial optimization based on REINFORCE
    Based on Kim et al. (2022) https://arxiv.org/abs/2205.13209

    Args:
        env: TorchRL Environment
        policy: Policy
        baseline: REINFORCE Baseline
        num_augment: Number of augmentations (default: 8)
        alpha: weight for invariance loss
        beta: weight for solution symmetricity loss
        augment_test: whether to augment data during testing as well
    """

    def __init__(
        self,
        env,
        policy=None,
        baseline=None,
        num_starts=10,
        num_augment=4,
        alpha=0.2,
        beta=1,
        augment_test=True,
        **policy_kwargs,
    ):
        super(SymNCO, self).__init__(env, policy, baseline)

        self.policy = (
            SymNCOPolicy(self.env, num_starts=num_starts, **policy_kwargs)
            if policy is None
            else policy
        )
        if baseline is not None:
            log.warn(
                "SymNCO uses shared baselines in the loss functions. Baseline argument will be ignored"
            )
        self.baseline = NoBaseline()  # baseline is calculated in the loss function

        # Multi-start parameters from policy, default to 1
        self.num_augment = num_augment
        self.augment = StateAugmentation(self.env.name)
        self.augment_test = augment_test
        self.alpha = alpha  # weight for invariance loss
        self.beta = beta  # weight for solution symmetricity loss

    def forward(self, td: TensorDict, phase: str = "train", extra=None, **policy_kwargs):
        """Evaluate model, get costs and log probabilities and compare with baseline"""

        # Get num_starts from policy. If single_traj, set num_starts and num_augment to 0
        num_starts = getattr(self.policy.decoder, "num_starts", 0)
        num_augment = self.num_augment

        if policy_kwargs.get("single_traj", False):
            num_starts, num_augment = 0, 0

        if num_augment > 1:
            td = self.augment(td, num_augment)

        # Evaluate model, get costs and log probabilities
        out = self.policy(td, phase, **policy_kwargs)

        # Unbatchify reward to [batch_size, num_starts, num_augment].
        reward = unbatchify(out["reward"], (num_starts, num_augment))

        # Get multi-start (=POMO) rewards and best actions
        if num_starts > 1:
            # max multi-start reward
            max_reward, max_idxs = reward.max(dim=1)
            out.update({"max_reward": max_reward})

            # Reshape batch to [batch, num_starts, num_augment]
            if out.get("actions", None) is not None:
                # TODO: actions are not unbatchified correctly
                actions = unbatchify(out["actions"], (num_starts, num_augment))
                out.update(
                    {"best_multistart_actions": gather_by_index(actions, max_idxs)}
                )
                out["actions"] = actions

        # Get augmentation score only during inference
        if num_augment > 1:
            # If multistart is enabled, we use the best multistart rewards
            reward_ = max_reward if num_starts > 1 else reward
            # [batch, num_augment]
            max_aug_reward, max_idxs = reward_.max(dim=1)
            out.update({"max_aug_reward": max_aug_reward})
            if out.get("best_multistart_actions", None) is not None:
                out.update(
                    {
                        "best_aug_actions": gather_by_index(
                            out["best_multistart_actions"], max_idxs
                        )
                    }
                )

        # Get best actions and rewards
        # Main training loss
        if phase == "train":
            # [batch_size, num_starts, num_augment]
            ll = unbatchify(out["log_likelihood"], (num_starts, num_augment))

            # Calculate losses: problem symmetricity, solution symmetricity, invariance

            loss_ps = problem_symmetricity_loss(reward, ll) if num_starts > 1 else 0
            loss_ss = solution_symmetricity_loss(reward, ll) if num_augment > 1 else 0
            loss_inv = (
                invariance_loss(out["proj_embeddings"], num_augment)
                if num_augment > 1
                else 0
            )
            loss = loss_ps + self.beta * loss_ss + self.alpha * loss_inv
            out.update(
                {
                    "loss": loss,
                    "loss_ss": loss_ss,
                    "loss_ps": loss_ps,
                    "loss_inv": loss_inv,
                }
            )

        return out



from rl4co.models.rl.reinforce.base import REINFORCE
from rl4co.models.rl.reinforce.baselines import RolloutBaseline, WarmupBaseline
from rl4co.models.zoo.ham.policy import HeterogeneousAttentionModelPolicy


class HeterogeneousAttentionModel(REINFORCE):
    """Heterogenous Attention Model for solving the Pickup and Delivery Problem based on REINFORCE
    https://arxiv.org/abs/2110.02634

    Args:
        env: TorchRL Environment
        policy: Policy
        baseline: REINFORCE Baseline
    """

    def __init__(self, env, policy=None, baseline=None, **policy_kwargs):
        super(HeterogeneousAttentionModel, self).__init__(env, policy, baseline)
        assert (
            self.env.name == "pdp"
        ), "HeterogeneousAttentionModel only works for PDP (Pickup and Delivery Problem)"
        self.policy = (
            HeterogeneousAttentionModelPolicy(self.env, **policy_kwargs)
            if policy is None
            else policy
        )
        self.baseline = (
            WarmupBaseline(RolloutBaseline()) if baseline is None else baseline
        )



from rl4co.models.rl.reinforce.base import REINFORCE
from rl4co.models.rl.reinforce.baselines import RolloutBaseline, WarmupBaseline
from rl4co.models.zoo.ptrnet.policy import PointerNetworkPolicy


class PointerNetwork(REINFORCE):
    """
    Pointer Network for neural combinatorial optimization based on REINFORCE
    Based on Vinyals et al. (2015) https://arxiv.org/abs/1506.03134
    Refactored from reference implementation: https://github.com/wouterkool/attention-learn-to-route

    Args:
        env: TorchRL Environment
        policy: Policy
        baseline: REINFORCE Baseline
    """

    def __init__(self, env, policy=None, baseline=None, **policy_kwargs):
        super(PointerNetwork, self).__init__(env, policy, baseline)
        self.policy = (
            PointerNetworkPolicy(self.env, **policy_kwargs) if policy is None else policy
        )
        self.baseline = (
            WarmupBaseline(RolloutBaseline()) if baseline is None else baseline
        )



# Base environment
# Main Environments
from rl4co.envs.atsp import ATSPEnv
from rl4co.envs.common.base import RL4COEnvBase
from rl4co.envs.cvrp import CVRPEnv
from rl4co.envs.dpp import DPPEnv
from rl4co.envs.mdpp import MDPPEnv
from rl4co.envs.mtsp import MTSPEnv
from rl4co.envs.op import OPEnv
from rl4co.envs.pctsp import PCTSPEnv
from rl4co.envs.pdp import PDPEnv
from rl4co.envs.sdvrp import SDVRPEnv
from rl4co.envs.spctsp import SPCTSPEnv
from rl4co.envs.tsp import TSPEnv



from typing import Any

import torch
import torch.nn as nn

from hydra.utils import instantiate
from lightning import LightningModule
from omegaconf import DictConfig
from torch.utils.data import DataLoader

from rl4co.data.dataset import tensordict_collate_fn
from rl4co.data.generate_data import generate_default_datasets
from rl4co.envs.common.base import RL4COEnvBase
from rl4co.utils.pylogger import get_pylogger

log = get_pylogger(__name__)


class RL4COLitModule(LightningModule):
    """
    Base LightningModule for Neural Combinatorial Optimization
    Args:
        cfg: Hydra config
        env: Environment to use overridding the config. If None, instantiate from config
        model: Model to use overridding the config. If None, instantiate from config
    """

    def __init__(
        self, cfg: DictConfig, env: RL4COEnvBase = None, model: nn.Module = None
    ):
        super().__init__()

        # this line ensures params passed to LightningModule will be saved to ckpt
        # it also allows to access params with 'self.hparams' attribute
        # self.save_hyperparameters("env", "model", logger=False)
        self.save_hyperparameters(logger=False)

        if cfg.get("train", {}).get("disable_profiling", True):
            # Disable profiling executor. This reduces memory and increases speed.
            # https://github.com/HazyResearch/safari/blob/111d2726e7e2b8d57726b7a8b932ad8a4b2ad660/train.py#LL124-L129C17
            try:
                torch._C._jit_set_profiling_executor(False)
                torch._C._jit_set_profiling_mode(False)
            except AttributeError:
                pass

        cfg = DictConfig(cfg) if not isinstance(cfg, DictConfig) else cfg
        self.cfg = cfg

        # Instantiate environment, model and metrics
        self.env = env if env is not None else self.instantiate_env()
        self.model = model if model is not None else self.instantiate_model()
        self.instantiate_metrics()

        if cfg.get("train", {}).get("manual_optimization", False):
            log.info("Manual optimization enabled")
            self.automatic_optimization = False

    def instantiate_env(self):
        log.info(f"Instantiating environment <{self.cfg.env._target_}>")
        return instantiate(self.cfg.env)

    def instantiate_model(self):
        log.info(f"Instantiating model <{self.cfg.model._target_}>")
        return instantiate(self.cfg.model, env=self.env)

    def instantiate_metrics(self):
        """Dictionary of metrics to be logged at each phase"""
        metrics = self.cfg.get("metrics", {})
        if not metrics:
            log.info("No metrics specified, using default")
        self.train_metrics = metrics.get("train", ["loss", "reward"])
        self.val_metrics = metrics.get("val", ["reward"])
        self.test_metrics = metrics.get("test", ["reward"])
        self.log_on_step = metrics.get("log_on_step", True)

    def setup(self, stage="fit"):
        log.info("Setting up batch sizes for train/val/test")
        # If any of the batch sizes are specified, use that. Otherwise, use the default batch size

        data_cfg = self.cfg.get("data", {})
        batch_size = data_cfg.get("batch_size", None)
        if data_cfg.get("train_batch_size", None) is not None:
            train_batch_size = data_cfg.train_batch_size
            if batch_size is not None:
                log.warning(
                    f"`train_batch_size`={train_batch_size} specified, ignoring `batch_size`={batch_size}"
                )
        elif batch_size is not None:
            train_batch_size = batch_size
        else:
            train_batch_size = 64
            log.warning(f"No batch size specified, using default as {train_batch_size}")
        # default all batch sizes to train_batch_size if not specified
        self.train_batch_size = train_batch_size
        self.val_batch_size = data_cfg.get("val_batch_size", train_batch_size)
        self.test_batch_size = data_cfg.get("test_batch_size", train_batch_size)

        log.info("Setting up datasets")

        # Create datasets automatically. If found, this will skip
        if data_cfg.get("generate_data", True):
            generate_default_datasets(
                data_dir=self.cfg.get("paths", {}).get("data_dir", "data/")
            )

        # If any of the dataset sizes are specified, use that. Otherwise, use the default dataset size
        def _get_phase_size(phase):
            DEFAULT_SIZES = {
                "train": 100000,
                "val": 10000,
                "test": 10000,
            }
            size = data_cfg.get(f"{phase}_size", None)
            if size is None:
                size = DEFAULT_SIZES[phase]
                message = f"No {phase}_size specified, using default as {size}"
                log.warning(message) if phase == "train" else log.info(message)
            return size

        self.train_size = _get_phase_size("train")
        self.val_size = _get_phase_size("val")
        self.test_size = _get_phase_size("test")
        self.train_dataset = self.wrap_dataset(self.env.dataset(self.train_size, "train"))
        self.val_dataset = self.env.dataset(self.val_size, "val")
        self.test_dataset = self.env.dataset(self.test_size, "test")

        if hasattr(self.model, "setup") and not self.cfg.get(
            "disable_model_setup", False
        ):
            self.model.setup(self)

    def configure_optimizers(self):
        train_cfg = self.cfg.get("train", {})
        if train_cfg.get("optimizer", None) is None:
            log.warning("No optimizer specified, using default")
        opt_cfg = train_cfg.get(
            "optimizer", DictConfig({"_target_": "torch.optim.Adam", "lr": 1e-4})
        )
        if "_target_" not in opt_cfg:
            log.info("No _target_ specified for optimizer, using default Adam")
            opt_cfg["_target_"] = "torch.optim.Adam"

        log.info(f"Instantiating optimizer <{opt_cfg._target_}>")
        optimizer = instantiate(opt_cfg, self.parameters())

        if "scheduler" not in train_cfg:
            return optimizer
        else:
            log.info(f"Instantiating scheduler <{train_cfg.scheduler._target_}>")
            lr_scheduler = instantiate(train_cfg.scheduler, optimizer)
            return [optimizer], {
                "scheduler": lr_scheduler,
                "interval": train_cfg.get("scheduler_interval", "epoch"),
                "monitor": train_cfg.get("scheduler_monitor", "val/reward"),
            }

    def shared_step(self, batch: Any, batch_idx: int, phase: str):
        td = self.env.reset(batch)
        out = self.model(td, phase, td.get("extra", None))

        # Log metrics
        metrics = getattr(self, f"{phase}_metrics")
        metrics = {f"{phase}/{k}": v.mean() for k, v in out.items() if k in metrics}

        log_on_step = self.log_on_step if phase == "train" else False
        on_epoch = False if phase == "train" else True
        self.log_dict(
            metrics,
            on_step=log_on_step,
            on_epoch=on_epoch,
            prog_bar=True,
            sync_dist=True,
            add_dataloader_idx=False,
        )
        return {"loss": out.get("loss", None), **metrics}

    def training_step(self, batch: Any, batch_idx: int):
        # To use new data every epoch, we need to call reload_dataloaders_every_epoch=True in Trainer
        return self.shared_step(batch, batch_idx, phase="train")

    def validation_step(self, batch: Any, batch_idx: int):
        return self.shared_step(batch, batch_idx, phase="val")

    def test_step(self, batch: Any, batch_idx: int):
        return self.shared_step(batch, batch_idx, phase="test")

    def train_dataloader(self):
        return self._dataloader(self.train_dataset, self.train_batch_size)

    def val_dataloader(self):
        return self._dataloader(self.val_dataset, self.val_batch_size)

    def test_dataloader(self):
        return self._dataloader(self.test_dataset, self.test_batch_size)

    def on_train_epoch_end(self):
        if hasattr(self.model, "on_train_epoch_end"):
            self.model.on_train_epoch_end(self)
        train_dataset = self.env.dataset(self.train_size, "train")
        self.train_dataset = self.wrap_dataset(train_dataset)

    def wrap_dataset(self, dataset):
        if hasattr(self.model, "wrap_dataset") and not self.cfg.get(
            "disable_wrap_dataset", False
        ):
            dataset = self.model.wrap_dataset(self, dataset)
        return dataset

    def _dataloader(self, dataset, batch_size):
        return DataLoader(
            dataset,
            batch_size=batch_size,
            shuffle=False,  # no need to shuffle, we're resampling every epoch
            num_workers=self.cfg.get("data", {}).get("num_workers", 0),
            collate_fn=tensordict_collate_fn,
        )



from rl4co.models.zoo.am import AttentionModel, AttentionModelPolicy
from rl4co.models.zoo.ham import (
    HeterogeneousAttentionModel,
    HeterogeneousAttentionModelPolicy,
)
from rl4co.models.zoo.mdam import MDAMPolicy
from rl4co.models.zoo.pomo import POMO, POMOPolicy
from rl4co.models.zoo.ptrnet import PointerNetwork, PointerNetworkPolicy
from rl4co.models.zoo.symnco import SymNCO, SymNCOPolicy



import os

import lightning as L
import torch
import yaml

from omegaconf import DictConfig

from rl4co.tasks.rl4co import RL4COLitModule
from rl4co.utils.pylogger import get_pylogger

log = get_pylogger(__name__)


def get_lightning_device(lit_module: L.LightningModule) -> torch.device:
    """Get the device of the Lightning module before setup is called
    See device setting issue in setup https://github.com/Lightning-AI/lightning/issues/2638
    """
    if lit_module.trainer.strategy.root_device != lit_module.device:
        return lit_module.trainer.strategy.root_device
    return lit_module.device


def remove_key(config, key="wandb"):
    """Remove keys containing 'key`"""
    new_config = {}
    for k, v in config.items():
        if key in k:
            continue
        else:
            new_config[k] = v
    return new_config


def clean_hydra_config(
    config, keep_value_only=True, remove_keys="wandb", clean_cfg_path=True
):
    """Clean hydra config by nesting dictionary and cleaning values"""
    # Remove keys containing `remove_keys`
    if not isinstance(remove_keys, list):
        remove_keys = [remove_keys]
    for key in remove_keys:
        config = remove_key(config, key=key)

    new_config = {}
    # Iterate over config dictionary
    for key, value in config.items():
        # If key contains slash, split it and create nested dictionary recursively
        if "/" in key:
            keys = key.split("/")
            d = new_config
            for k in keys[:-1]:
                d = d.setdefault(k, {})
            d[keys[-1]] = value["value"] if keep_value_only else value
        else:
            new_config[key] = value["value"] if keep_value_only else value

    cfg = DictConfig(new_config)

    if clean_cfg_path:
        # Clean cfg_path recursively substituting root_dir with cwd
        root_dir = cfg.paths.root_dir

        def replace_dir_recursive(d, search, replace):
            for k, v in d.items():
                if isinstance(v, dict) or isinstance(v, DictConfig):
                    replace_dir_recursive(v, search, replace)
                elif isinstance(v, str):
                    if search in v:
                        d[k] = v.replace(search, replace)

        replace_dir_recursive(cfg, root_dir, os.getcwd())

    return cfg


def load_model_from_checkpoint(
    config,
    checkpoint_path,
    device="cpu",
    only_policy=True,
    disable_model_setup=True,
    disable_wrap_dataset=True,
    validate_only=True,
    clean_cfg_path=True,
    phase="test",
):
    """Load model from checkpoint

    Args:
        config: Hydra config or its path
        checkpoint_path: Path to checkpoint
        device: Device to load model on
        only_policy: If True, load only policy parameters
        disable_model_setup: If True, disable model setup during RL4COLitModule init
        disable_wrap_dataset: If True, disable dataset wrapping during RL4COLitModule init
        validate_only: If True, only load model for validation and make train size small
    """
    if only_policy and not (disable_model_setup or disable_wrap_dataset):
        log.warning(
            "only_policy is True, but disable_model_setup and disable_wrap_dataset are False. "
            "This may cause errors due to missing model setup and dataset wrapping. "
        )

    # Load config if path is given
    if not isinstance(config, DictConfig or dict):
        log.info(f"Loading config from {config}")
        with open(config, "r") as stream:
            config = yaml.safe_load(stream)

    # Clean hydra config
    config = clean_hydra_config(config, clean_cfg_path=clean_cfg_path)

    # Add to cfg disable_model_setup and disable_wrap_dataset
    config["disable_model_setup"] = disable_model_setup
    config["disable_wrap_dataset"] = disable_wrap_dataset
    if validate_only:
        config["train_size"] = 10  # dummy

    # Load model and checkpoint
    lit_module = RL4COLitModule(config)
    checkpoint_path = torch.load(checkpoint_path, map_location=device)

    # Load model from checkpoint: only policy parameters or full model
    if only_policy:
        state_dict = checkpoint_path["state_dict"]
        # get only policy parameters
        state_dict = {k: v for k, v in state_dict.items() if "policy" in k}
        # remove leading 'policy.' from keys
        state_dict = {k.replace("model.policy.", ""): v for k, v in state_dict.items()}
        # load policy state_dict
        lit_module.model.policy.load_state_dict(state_dict)
    else:
        lit_module = lit_module.load_from_checkpoint(checkpoint_path)

    lit_module.setup(stage=phase)
    return lit_module



from torch import nn

from rl4co.models.nn.graph.gat import GraphAttentionEncoder


class CriticNetwork(nn.Module):
    """We make the critic network compatible with any problem by using encoder for any environment
    Refactored from Kool et al. (2019) which only worked for TSP
    Reference: https://github.com/wouterkool/attention-learn-to-route

    Args:
        env (EnvBase): environment
        encoder (nn.Module, optional): encoder. Defaults to None. Initialized with GraphAttentionEncoder.
        embedding_dim (int, optional): embedding dimension. Defaults to 128.
        hidden_dim (int, optional): hidden dimension. Defaults to 512.
        n_layers (int, optional): number of encoder layers. Defaults to 3.
        num_heads (int, optional): number of attention heads. Defaults to 8.
        encoder_normalization (str, optional): normalization. Defaults to "batch".
    """

    def __init__(
        self,
        env=None,
        encoder=None,
        embedding_dim=128,
        hidden_dim=512,
        num_layers=3,
        num_heads=8,
        encoder_normalization="batch",
        use_native_sdpa=False,
        force_flash_attn=False,
    ):
        super(CriticNetwork, self).__init__()

        self.encoder = (
            GraphAttentionEncoder(
                num_heads=num_heads,
                embedding_dim=embedding_dim,
                num_layers=num_layers,
                env=env,
                normalization=encoder_normalization,
                feed_forward_hidden=hidden_dim,
                use_native_sdpa=use_native_sdpa,
                force_flash_attn=force_flash_attn,
            )
            if encoder is None
            else encoder
        )

        self.value_head = nn.Sequential(
            nn.Linear(embedding_dim, hidden_dim), nn.ReLU(), nn.Linear(hidden_dim, 1)
        )

    def forward(self, td):
        graph_embeddings, _ = self.encoder(td)
        # graph_embedings: [batch_size, graph_size, input_dim]
        # return self.value_head(graph_embeddings.mean(1))

        # L2D style
        return self.value_head(graph_embeddings).mean(1)



"""Basic utilities for common tasks in Python and PyTorch."""
import re

from pathlib import Path

import torch


def flatten_params(params):
    """Flatten an iterable of parameters."""
    flat_params = [p.contiguous().view(-1) for p in params]
    return torch.cat(flat_params) if len(flat_params) > 0 else torch.tensor([])


def flatten_params_grad(params, params_ref):
    """Flatten an iterable of parameters and their gradients."""
    _params = [p for p in params]
    _params_ref = [p for p in params_ref]
    flat_params = [
        p.contiguous().view(-1) if p is not None else torch.zeros_like(q).view(-1)
        for p, q in zip(_params, _params_ref)
    ]
    return torch.cat(flat_params) if len(flat_params) > 0 else torch.tensor([])


def parameter_count(model):
    "Returns parameter count of an nn.Module."
    return sum([p.numel() for p in model.parameters()])


def strictly_increasing(L):
    return all(x < y for x, y in zip(L, L[1:]))


def strictly_decreasing(L):
    return all(x > y for x, y in zip(L, L[1:]))


def non_increasing(L):
    return all(x >= y for x, y in zip(L, L[1:]))


def non_decreasing(L):
    return all(x <= y for x, y in zip(L, L[1:]))


def monotonic(L):
    return non_increasing(L) or non_decreasing(L)


def find(tensor, values):
    "Finds indices of elements in a tensor that are equal to values."
    return torch.nonzero(tensor[..., None] == values)


def sum_except(x, num_dims=1):
    """
    Sums all dimensions except the first `num_dims`.
    Args:
        x: Tensor, shape (batch_size, ...)
        num_dims: int, number of batch dims (default=1)
    Returns:
        x_sum: Tensor, shape (batch_size,)
    """
    return x.reshape(*x.shape[:num_dims], -1).sum(-1)


def load_checkpoint(path, device="cpu"):
    "Loads nn.Module from a path."
    path = Path(path).expanduser()
    is_deepspeed = False
    if path.is_dir():  # DeepSpeed checkpoint
        is_deepspeed = True
        latest_path = path / "latest"
        if latest_path.is_file():
            with open(latest_path, "r") as fd:
                tag = fd.read().strip()
        else:
            raise ValueError(f"Unable to find 'latest' file at {latest_path}")
        path /= f"{tag}/mp_rank_00_model_states.pt"
    state_dict = torch.load(path, map_location=device)
    if is_deepspeed:
        state_dict = state_dict["module"]

        # Replace the names of some of the submodules
        def key_mapping(key):
            return re.sub(r"^module.model.", "", key)

        state_dict = {key_mapping(k): v for k, v in state_dict.items()}
    return state_dict



import pytest

from rl4co.models import (
    POMO,
    AttentionModel,
    HeterogeneousAttentionModel,
    MDAMPolicy,
    PointerNetwork,
    SymNCO,
    SymNCOPolicy,
)
from rl4co.utils.test_utils import generate_env_data


@pytest.mark.parametrize("size", [20])
@pytest.mark.parametrize(
    "env_name", ["tsp", "cvrp", "sdvrp", "mtsp", "op", "pctsp", "spctsp", "dpp", "mdpp"]
)  # todo: sdvrp
def test_am(size, env_name, batch_size=2):
    env, x = generate_env_data(env_name, size, batch_size)
    td = env.reset(x)
    model = AttentionModel(env)
    out = model(td, decode_type="sampling")
    assert out["reward"].shape == (batch_size,)


@pytest.mark.parametrize("size", [20])
def test_ptrnet(size, batch_size=2):
    env, x = generate_env_data("tsp", size, batch_size)
    td = env.reset(x)
    model = PointerNetwork(env)
    out = model(td, decode_type="sampling")
    assert out["reward"].shape == (batch_size,)


@pytest.mark.parametrize("size", [20])
def test_pomo(size, batch_size=2):
    env, x = generate_env_data("tsp", size, batch_size)
    td = env.reset(x)
    model = POMO(env, num_starts=size)
    out = model(td, decode_type="sampling")
    assert out["reward"].shape == (batch_size * size,)


@pytest.mark.parametrize("size", [20])
def test_symnco(size, batch_size=2, num_augment=8, num_starts=20):
    env, x = generate_env_data("tsp", size, batch_size)
    td = env.reset(x)
    policy = SymNCOPolicy(env, num_starts=num_starts)
    model = SymNCO(env, policy, num_augment=num_augment)
    out = model(td, decode_type="sampling")
    assert out["reward"].shape == (batch_size * num_augment * num_starts,)


@pytest.mark.parametrize("size", [20])
def test_ham(size, batch_size=2):
    env, x = generate_env_data("pdp", size, batch_size)
    td = env.reset(x)
    model = HeterogeneousAttentionModel(env)
    out = model(td, decode_type="sampling")
    assert out["reward"].shape == (batch_size,)


@pytest.mark.parametrize("size", [20])
def test_mdam(size, batch_size=2, num_paths=5):
    env, x = generate_env_data("tsp", size, batch_size)
    td = env.reset(x)
    model = MDAMPolicy(env, num_paths=num_paths)
    out = model(td, decode_type="sampling")
    print(out["reward"].shape)
    assert out["reward"].shape == (
        num_paths,
        batch_size,
    )
